package naito.extaction;

import adf.agent.action.ambulance.ActionLoad;
import adf.agent.action.ambulance.ActionRescue;
import adf.agent.action.ambulance.ActionUnload;
import adf.agent.action.common.ActionMove;
import adf.agent.info.AgentInfo;
import adf.agent.info.ScenarioInfo;
import adf.agent.info.WorldInfo;
import adf.agent.module.ModuleManager;
import adf.component.extaction.ExtAction;
import adf.component.module.algorithm.PathPlanning;
import naito.SampleModuleKey;
import naito.util.DistanceSorter;
import rescuecore2.standard.entities.Area;
import rescuecore2.standard.entities.Human;
import rescuecore2.standard.entities.StandardEntityURN;
import rescuecore2.worldmodel.EntityID;

import java.util.*;


public class KKActionRescue extends ExtAction {
    EntityID target;

    public KKActionRescue(AgentInfo ai, WorldInfo wi, ScenarioInfo si, ModuleManager moduleManager) {
        super(ai, wi, si, moduleManager);
        this.target = null;
    }

    @Override
    public ExtAction setTarget(EntityID... entityIDs) {
        Area area;
        Human victim = Arrays.asList(entityIDs)
                .stream()
                .filter((a) -> worldInfo.getEntity(a) instanceof Human)
                .map((a) -> (Human)worldInfo.getEntity(a))
                .min(new DistanceSorter(worldInfo, this.agentInfo.me()))
                .orElse(null);

        if(victim == null) {
            area = Arrays.asList(entityIDs)
                    .stream().filter((a) -> worldInfo.getEntity(a) instanceof Area)
                    .map(a -> (Area)worldInfo.getEntity(a))
                    .min(new DistanceSorter(worldInfo, this.agentInfo.me()))
                    .orElse(null);
            target = (area != null) ? area.getID() : null;
        }
        else {
            target = victim.getID();
        }

        if (scenarioInfo.isDebugMode())
            System.err.printf("( %d) %d %s\n", agentInfo.getTime(), agentInfo.getID().getValue(), "Target = " + ((target != null) ? target.getValue() : "null"));

        return this;
    }

    private boolean isBuriedSatisfied(Human victim)
    {
        return (!victim.isBuriednessDefined() || (victim.getBuriedness() > 0));
    }

    @Override
    public ExtAction calc() {
        this.result = null;
        if(this.target == null) {
            return this;
        }

        if(this.worldInfo.getEntity(target) instanceof Area) {
            PathPlanning pp = this.moduleManager.getModule(SampleModuleKey.AMBULANCE_MODULE_PATH_PLANNING);
            pp.setFrom(this.agentInfo.getPosition()).setDestination(target);
            List<EntityID> path = pp.calc().getResult();
            if(path != null) {
                this.result = new ActionMove(path);
                if (scenarioInfo.isDebugMode())
                    System.err.printf("( %d) %d %s\n", agentInfo.getTime(), agentInfo.getID().getValue(), "Move area to rescue search : " + path + " from " + agentInfo.getPosition().getValue());
            }

            return this;
        }

        Human victim = (Human)this.worldInfo.getEntity(target);
        Human me = (Human)this.agentInfo.me();

        if(agentInfo.someoneOnBoard() != null) {
            // Loading
            if(agentInfo.getPositionArea().getStandardURN() == StandardEntityURN.REFUGE)
            {
                // Unload
                this.result = new ActionUnload();
                if (scenarioInfo.isDebugMode()) System.err.printf("( %d) %d %s\n", agentInfo.getTime(), agentInfo.getID().getValue(), "Unload");
            }
            else {
                // Move to Refuge
                PathPlanning pp = this.moduleManager.getModule(SampleModuleKey.AMBULANCE_MODULE_PATH_PLANNING);
                pp.setFrom(agentInfo.getPosition())
                        .setDestination(worldInfo.getEntityIDsOfType(StandardEntityURN.REFUGE));
                List<EntityID> path = pp.calc().getResult();
                if (path != null) {
                    this.result = new ActionMove(path);
                    if (scenarioInfo.isDebugMode())
                        System.err.printf("( %d) %d %s\n", agentInfo.getTime(), agentInfo.getID().getValue(), "Move Area to unload : " + path + " from " + agentInfo.getPosition().getValue());
                }
            }
        }
        else {
            // Rescuing
            if(victim.getPosition().getValue() == me.getPosition().getValue()) {
                // Rescue or Load
                if(isBuriedSatisfied(victim)) {
                    this.result = new ActionRescue(target);
                    if (scenarioInfo.isDebugMode())
                        System.err.printf("( %d) %d %s\n", agentInfo.getTime(), agentInfo.getID().getValue(), "Rescuing " + this.target.getValue() + "(B: " + (victim.isBuriednessDefined() ? victim.getBuriedness() : "null") + ")");
                }
                else if(victim.getStandardURN() == StandardEntityURN.CIVILIAN
                        && worldInfo.getPosition(victim).getStandardURN() != StandardEntityURN.REFUGE
                       ) {
                    this.result = new ActionLoad(target);
                    if (scenarioInfo.isDebugMode()) System.err.printf("( %d) %d %s\n", agentInfo.getTime(), agentInfo.getID().getValue(), "Loading " + this.target.getValue() +"(D: " + (victim.isDamageDefined() ? victim.getDamage() : "null") + ")");
                }
                else {
                    if (scenarioInfo.isDebugMode()) System.err.printf("( %d) %d %s\n", agentInfo.getTime(), agentInfo.getID().getValue(), "Nothing (D: " + (victim.isDamageDefined() ? victim.getDamage() : "null") + ", B: " + (victim.isBuriednessDefined() ? victim.getBuriedness() : "null") + ")");
                }
            }
            else {
                // Move to
                if (victim.getPosition().getValue() != this.agentInfo.getPosition().getValue()) { // Position Check
                    PathPlanning pp = this.moduleManager.getModule(SampleModuleKey.AMBULANCE_MODULE_PATH_PLANNING);
                    pp.setFrom(agentInfo.getPosition()).setDestination(victim.getPosition());
                    List<EntityID> path = pp.calc().getResult();
                    if (path != null) {
                        int x = -1;
                        int y = -1;
                        if (victim.isXDefined() && victim.isYDefined()) {
                            x = victim.getX();
                            y = victim.getY();
                        }
                        this.result = new ActionMove(path, x, y);
                        if (scenarioInfo.isDebugMode()) System.err.printf("( %d) %d %s\n", agentInfo.getTime(), agentInfo.getID().getValue(), "Move to Rescue Agent : " + path + " from " + agentInfo.getPosition().getValue() + " to " + victim.getPosition().getValue());
                    }
                }
            }
        }

        return this;
    }
}
