package naito.extaction.mactionextclear;

import adf.agent.action.Action;
import adf.agent.action.common.ActionMove;
import adf.agent.action.common.ActionRest;
import adf.agent.action.police.ActionClear;
import adf.agent.info.AgentInfo;
import adf.agent.info.ScenarioInfo;
import adf.agent.info.WorldInfo;
import adf.agent.module.ModuleManager;
import adf.component.extaction.ExtAction;
import adf.component.module.algorithm.Clustering;
import adf.component.module.algorithm.PathPlanning;
import naito.SampleModuleKey;
import naito.util.passable.CheckAreaPassable;
import naito.util.passable.SearchPoliceClearLine;
import rescuecore2.misc.geometry.GeometryTools2D;
import rescuecore2.misc.geometry.Line2D;
import rescuecore2.misc.geometry.Point2D;
import rescuecore2.standard.entities.*;
import rescuecore2.worldmodel.EntityID;

import java.util.*;

public class MActionExtClear extends ExtAction
{
    private EntityID       target;
    private List<EntityID> path;
    private Integer        pathIndex;

    private static final double AGENT_SIZE = 500;

    public MActionExtClear(
        AgentInfo     ai,
        WorldInfo     wi,
        ScenarioInfo  si,
        ModuleManager mm
    ){
        super(ai, wi, si, mm);
    }

    @Override
    public ExtAction setTarget(EntityID... targets)
    {
        if(targets != null)
            this.target = targets[0];

        return this;
    }

    @Override
    public ExtAction calc()
    {
        this.result = null;

        if (this.target == null)
            return this;

        if (!this.target.equals(Util.getLast(this.path)) || !this.path.contains(this.agentInfo.getPosition()))
        {
            this.path = this.calculatePath(this.target);
            this.pathIndex = 0;
        }

        if (this.path == null)
        {
            this.result = (this.agentInfo.getPosition().equals(this.target)) ? new ActionRest() : null;
            return this;
        }

        if (this.path.size() == 0)
        {
            this.result = (this.agentInfo.getPosition().equals(this.target)) ? new ActionRest() : null;
            return this;
        }

        this.pathIndex = this.findNextClearArea(this.path, this.pathIndex);
        EntityID trueTarget = this.path.get(this.pathIndex);

        if (trueTarget == null)
        {
            this.result = (this.agentInfo.getPosition().equals(this.target)) ? new ActionRest() : null;
            return this;
        }

        System.out.println(trueTarget);
        Line2D clearLine = this.decideClearLine(this.path, this.pathIndex);

        if (clearLine == null)
        {
            this.result = (trueTarget.equals(this.agentInfo.getPosition()))
                ?   new ActionRest()
                :   this.createActionMove(trueTarget, null, this.path);

            return this;
        }

        double distance = GeometryTools2D.getDistance(
            Util.toPoint2D(this.agentInfo),
            clearLine.getOrigin()
        );

        this.result = (distance > AGENT_SIZE * 3)
            ?   this.createActionMove(trueTarget, clearLine.getOrigin(), this.path)
            :   this.createActionClear(clearLine.getEndPoint());

        if (this.result == null && this.agentInfo.getPosition().equals(this.target))
            this.result = new ActionRest();

        return this;
    }

    private List<EntityID> calculatePath(EntityID destination)
    {
        PathPlanning pathPlanning =
            this.moduleManager.getModule(SampleModuleKey.POLICE_MODULE_PATH_PLANNING);

        pathPlanning.setFrom(this.agentInfo.getPosition()).setDestination(destination);
        List<EntityID> path = pathPlanning.calc().getResult();
        path.add(0, this.agentInfo.getPosition());

        return path;
    }

    private int findNextClearArea(List<EntityID> path, int index)
    {
        Clustering areaIsUpdatedInfo = this.moduleManager.getModule(
            "",
            "naito.module.algorithm.MAreaIsUpdatedInfo"
        );
        Set<EntityID> areaIsUpdatedInfoSet = new HashSet<>(
            areaIsUpdatedInfo.getClusterEntityIDs(0)
        );

        for (; index<path.size(); index++)
        {
            EntityID entityID = path.get(index);
            StandardEntity entity = this.worldInfo.getEntity(entityID);

            if (index == path.size()-1)
                break;
            if (entity instanceof Building)
                continue;
            if (!(entity instanceof Area))
                break;
            if (!areaIsUpdatedInfoSet.contains(entityID))
                break;

            //if (this.findHumanInBlockade(entityID).size() != 0)
            //    break;

            Area area = (Area)entity;
            StandardEntity nextEntity =
                this.worldInfo.getEntity(path.get(index+1));

            if (!(nextEntity instanceof Area)) break;

            Area nextArea = (Area)nextEntity;

            CheckAreaPassable checker = new CheckAreaPassable();
            boolean checkResult;

            if (entityID.equals(this.agentInfo.getPosition()))
            {
                Point2D agentPoint = Util.toPoint2D(this.agentInfo);
                checkResult =
                    checker.check(area, agentPoint, nextArea, this.worldInfo);
            }
            else
            {
                StandardEntity backEntity =
                    this.worldInfo.getEntity(path.get(index-1));

                if (!(backEntity instanceof Area)) break;

                Area backArea = (Area)backEntity;

                checkResult =
                    checker.check(area, backArea, nextArea, this.worldInfo);
            }

            if (checkResult == false) break;
        }

        return index;
    }

    private Line2D decideClearLine(List<EntityID> path, int index)
    {
        EntityID areaID = path.get(index);

        EntityID entranceAreaID = (index-1 >= 0) ? path.get(index-1) : null;
        EntityID exitAreaID = (index+1 < path.size()) ? path.get(index+1) : null;

        List<List<Line2D>> searchResult =
            this.searchCandidateOfClearLine(areaID, entranceAreaID, exitAreaID);

        if (searchResult == null)
            return null;

        double minDistance = Double.MAX_VALUE;
        Line2D closestLine = null;

        for (List<Line2D> clearCurve : searchResult)
        {
            if (clearCurve.size() == 0) continue;

            Line2D  beginLine  = clearCurve.get(0);
            Point2D beginPoint = beginLine.getOrigin();
            double  beginsDistance =
                GeometryTools2D.getDistance(Util.toPoint2D(this.agentInfo), beginPoint);

            if (closestLine == null || beginsDistance < minDistance)
            {
                minDistance = beginsDistance;
                closestLine = beginLine;
            }
        }

        for (Point2D humanPoint : this.findHumanInBlockade(areaID))
        {
            double distance =
                GeometryTools2D.getDistance(Util.toPoint2D(this.agentInfo), humanPoint);
            if (humanPoint == null || distance < minDistance)
            {
                minDistance = distance;
                closestLine = new Line2D(humanPoint, humanPoint);
            }
        }


        return closestLine;
    }

    private List<Point2D> findHumanInBlockade(EntityID areaID)
    {
        List<StandardEntity> allEntities = new ArrayList<>(
            this.worldInfo.getAllEntities()
        );

        List<Point2D> humanPointsInBlockade = new ArrayList<>();

        StandardEntity areaEntity = this.worldInfo.getEntity(areaID);
        if (!(areaEntity instanceof Area))
            return humanPointsInBlockade;

        Area area = (Area)areaEntity;
        List<java.awt.Shape> blockadeShapes =
            Util.toShapes(Util.toBlockades(area, this.worldInfo));

        for (StandardEntity entity : allEntities)
        {
            if (!(entity instanceof Human))
                continue;

            if (entity.getID().equals(this.agentInfo.getID()))
                continue;

            Human human = (Human)entity;
            if (!human.getPosition().equals(areaID))
                continue;

            if (!human.isXDefined() || !human.isYDefined())
                continue;

            for (java.awt.Shape blockadeShape : blockadeShapes)
            {
                if (blockadeShape.contains(human.getX(), human.getY()))
                {
                    humanPointsInBlockade.add(new Point2D(human.getX(), human.getY()));
                    break;
                }
            }
        }

        return humanPointsInBlockade;
    }

    private List<List<Line2D>> searchCandidateOfClearLine(
        EntityID areaID,
        EntityID entranceAreaID,
        EntityID exitAreaID
    ){
        StandardEntity entity = this.worldInfo.getEntity(areaID);

        if (entity instanceof Building)
            return null;

        if (!(entity instanceof Area))
            return null;

        Area area = (Area)entity;

        Area    entranceArea = null;
        Area    exitArea     = null;
        Point2D beginPoint   = null;
        Line2D  exitLine     = null;

        if (entranceAreaID != null)
        {
            StandardEntity entranceEntity =
                this.worldInfo.getEntity(entranceAreaID);

            if (!(entranceEntity instanceof Area))
                return null;

            entranceArea = (Area)entranceEntity;
        }

        if (exitAreaID != null)
        {
            StandardEntity exitEntity =
                this.worldInfo.getEntity(exitAreaID);

            if (!(exitEntity instanceof Area))
                return null;

            exitArea = (Area)exitEntity;
        }

        if (areaID.equals(this.agentInfo.getPosition()))
        {
            beginPoint = Util.toPoint2D(this.agentInfo);
        }

        if (entranceAreaID != null || exitAreaID == null)
        {
            Edge duplicateEdge =
                Util.findDuplicateEdge(area, entranceArea);

            exitLine =
                Util.findFarthest(Util.toLines(area), duplicateEdge.getLine());
        }

        SearchPoliceClearLine searcher = new SearchPoliceClearLine();

        if (beginPoint != null && exitArea != null)
            return searcher.search(area, beginPoint, exitArea, this.worldInfo);

        if (beginPoint != null && exitLine != null)
            return searcher.search(area, beginPoint, exitLine, this.worldInfo);

        if (entranceArea != null && exitArea != null)
            return searcher.search(area, entranceArea, exitArea, this.worldInfo);

        if (entranceArea != null && exitLine != null)
            return searcher.search(area, entranceArea, exitLine, this.worldInfo);

        return null;
    }

    private Action createActionMove(EntityID target, Point2D point, List<EntityID> clearPath)
    {
        List<EntityID> movePath = new ArrayList<>();

        int i;
        for (i=0; i<clearPath.size(); i++)
        {
            if (clearPath.get(i).equals(this.agentInfo.getPosition()))
                break;
        }

        for (; i<clearPath.size(); i++)
        {
            movePath.add(clearPath.get(i));

            if (clearPath.get(i).equals(target))
                break;
        }

        return (point == null)
            ?   new ActionMove(movePath)
            :   new ActionMove(movePath, (int)point.getX(), (int)point.getY());
    }

    private Action createActionClear(Point2D point)
    {
        Line2D baseClearLine = new Line2D(Util.toPoint2D(this.agentInfo), point);
        Line2D extendedClearLine = new Line2D(
            Util.toPoint2D(this.agentInfo),
            baseClearLine.getDirection().normalised().scale(1000000)
        );
        Point2D clearPoint = extendedClearLine.getEndPoint();

        return new ActionClear((int)clearPoint.getX(), (int)clearPoint.getY());
    }
}
