package naito.extaction.mactionextclear;

import rescuecore2.misc.geometry.*;
import rescuecore2.worldmodel.EntityID;
import rescuecore2.standard.entities.Area;
import rescuecore2.standard.entities.Edge;
import adf.agent.info.AgentInfo;
import adf.agent.info.WorldInfo;
import rescuecore2.standard.entities.*;

import java.util.*;
import java.util.stream.*;

public class Util
{
    public static Point2D toPoint2D(AgentInfo ai)
    {
        return new Point2D(ai.getX(), ai.getY());
    }

    public static Point2D toPoint2D(Area area)
    {
        return new Point2D(area.getX(), area.getY());
    }

    public static <T> T getLast(List<T> list)
    {
        if (list == null)
            return null;

        if (list.size() == 0)
            return null;

        return list.get(list.size()-1);
    }

    public static boolean isEqual(Line2D line1, Line2D line2)
    {
        Point2D origin1 = line1.getOrigin();
        Point2D origin2 = line2.getOrigin();
        Point2D end1 = line1.getEndPoint();
        Point2D end2 = line2.getEndPoint();
        if (origin1.equals(origin2) && end1.equals(end2))
            return true;

        if (origin1.equals(end2) && origin2.equals(end1))
            return true;

        return false;
    }

    public static Line2D findFarthest(List<Line2D> candidates, Line2D base)
    {
        Line2D farthest = null;
        double maxDistance = 0.0;

        for (Line2D candidate : candidates)
        {
            if (isEqual(base, candidate)) continue;

            double distance = GeometryTools2D.getDistance(
                getCenter(candidate),
                getCenter(base)
            );

            if (distance > maxDistance)
            {
                farthest = candidate;
                maxDistance = distance;
            }
        }

        return farthest;
    }

    public static Point2D getCenter(Line2D line)
    {
        Point2D origin   = line.getOrigin();
        Point2D endpoint = line.getEndPoint();

        return new Point2D(
            origin.getX() + (endpoint.getX()-origin.getX())/2,
            origin.getY() + (endpoint.getY()-origin.getY())/2
        );
    }

    public static List<Line2D> toLines(Area area)
    {
        return area.getEdges().stream()
            .map(Edge::getLine)
            .collect(Collectors.toList());
    }

    public static Edge findDuplicateEdge(Area area1, Area area2)
    {
        List<Edge> edges1 = area1.getEdges().stream()
            .filter(Edge::isPassable)
            .collect(Collectors.toList());

        List<Edge> edges2 = area2.getEdges().stream()
            .filter(Edge::isPassable)
            .collect(Collectors.toList());

        for (Edge edge1 : edges1) for (Edge edge2 : edges2)
        {
            Line2D line1 = edge1.getLine();
            Line2D line2 = edge2.getLine();
            if (isEqual(line1, line2))
                return edge1;
        }

        return null;
    }

    public static List<Blockade> toBlockades(Area area, WorldInfo wi)
    {
        if (area.isBlockadesDefined() == false)
            return new ArrayList<Blockade>();

        List<EntityID> blockadeIDs = area.getBlockades();
        if (blockadeIDs == null)
            return new ArrayList<Blockade>();

        return blockadeIDs.stream()
            .map(eid -> (Blockade)wi.getEntity(eid))
            .collect(Collectors.toList());
    }

    public static List<java.awt.Shape> toShapes(List<Blockade> blockades)
    {
        return blockades.stream()
            .map(Blockade::getShape)
            .collect( Collectors.toList() );
    }
}
