package naito.module.algorithm;

import adf.agent.info.AgentInfo;
import adf.agent.info.ScenarioInfo;
import adf.agent.info.WorldInfo;
import adf.agent.module.ModuleManager;
import adf.agent.precompute.PrecomputeData;
import adf.agent.communication.MessageManager;
import adf.component.module.algorithm.Clustering;
import adf.component.module.algorithm.DynamicClustering;
import adf.util.WorldUtil;
import rescuecore2.misc.Pair;
import rescuecore2.misc.collections.LazyMap;
import rescuecore2.misc.geometry.Point2D;
import rescuecore2.standard.entities.*;
import rescuecore2.worldmodel.Entity;
import rescuecore2.worldmodel.EntityID;

import java.util.*;
import java.util.stream.*;

import static java.util.Comparator.comparing;
import static java.util.Comparator.reverseOrder;

public class MAreaIsUpdatedInfo extends DynamicClustering
{
    private Set<EntityID> updatedAreaIDs;

    public MAreaIsUpdatedInfo(
        AgentInfo     ai,
        WorldInfo     wi,
        ScenarioInfo  si,
        ModuleManager moduleManager
    ){
        super(ai, wi, si, moduleManager);
        this.updatedAreaIDs = new HashSet<>();
    }

    @Override
    public int getClusterNumber()
    {
        return 1;
    }

    @Override
    public int getClusterIndex(StandardEntity entity)
    {
        return this.getClusterIndex(entity.getID());
    }

    @Override
    public int getClusterIndex(EntityID id)
    {
        if (this.updatedAreaIDs.contains(id))
            return 0;

        return -1;
    }

    @Override
    public Collection<StandardEntity> getClusterEntities(int index)
    {
        if (index != 0) return null;

        return this.updatedAreaIDs.stream()
            .map(e -> this.worldInfo.getEntity(e))
            .collect(Collectors.toList());
    }

    @Override
    public Collection<EntityID> getClusterEntityIDs(int index)
    {
        if (index != 0) return null;

        return new HashSet<>(this.updatedAreaIDs);
    }

    @Override
    public DynamicClustering calc()
    {
        return this;
    }

    @Override
    public Clustering updateInfo(MessageManager messageManager)
    {
        super.updateInfo(messageManager);

        for (EntityID entityID : this.worldInfo.getChanged().getChangedEntities())
        {
            this.updatedAreaIDs.add(entityID);
        }

        return this;
    }
}
