package naito.module.algorithm;

import adf.agent.communication.MessageManager;
import adf.agent.info.AgentInfo;
import adf.agent.info.ScenarioInfo;
import adf.agent.info.WorldInfo;
import adf.agent.module.ModuleManager;
import adf.agent.precompute.PrecomputeData;
import adf.component.module.algorithm.PathPlanning;
import naito.util.passable.CheckAreaPassable;
import rescuecore2.misc.collections.LazyMap;
import rescuecore2.standard.entities.Area;
import rescuecore2.standard.entities.Blockade;
import rescuecore2.worldmodel.Entity;
import rescuecore2.worldmodel.EntityID;

import java.util.*;

public class SamplePlusPathPlanning extends PathPlanning
{
    /*
    private class ImpassableAreaMemory
    {
        private ArrayList<ArrayList<EntityID>> list = new ArrayList<>();
        private int pointer, size;
        public ImpassableAreaMemory(int size)
        {
            this.pointer = 0;
            this.size = size;
            for (int i = 0; i < size; i++)
            {
                list.add(new ArrayList<>());
            }
        }

        public void forward()
        {
            if (pointer < size)
            { pointer++; }
            else
            { pointer = 0; }
            list.get(pointer).clear();
        }

        public boolean contains(EntityID id)
        {
            for (ArrayList l : list)
            {
                if (l.contains(id))
                { return true; }
            }
            return false;
        }

        public void put(EntityID id)
        {
            list.get(pointer).add(id);
        }

        public void remove(EntityID id)
        {
            for (ArrayList l : list)
            {
                l.remove(id);
            }
        }
    }

    private ImpassableAreaMemory impassableAreaMemory = new ImpassableAreaMemory(20);
    */
    private CheckAreaPassable checkAreaPassable = new CheckAreaPassable();

    private Map<EntityID, Set<EntityID>> graph;

    private EntityID from;
    private Collection<EntityID> targets;
    private List<EntityID> result;

    public SamplePlusPathPlanning(AgentInfo ai, WorldInfo wi, ScenarioInfo si, ModuleManager moduleManager)
    {
        super(ai, wi, si, moduleManager);
        this.init();
    }

    private void init()
    {
        Map<EntityID, Set<EntityID>> neighbours = new LazyMap<EntityID, Set<EntityID>>()
        {
            @Override
            public Set<EntityID> createValue() {
                return new HashSet<>();
            }
        };
        for (Entity next : this.worldInfo)
        {
            if (next instanceof Area)
            {
                Collection<EntityID> areaNeighbours = ((Area) next).getNeighbours();
                neighbours.get(next.getID()).addAll(areaNeighbours);
            }
        }
        this.graph = neighbours;
    }

    @Override
    public List<EntityID> getResult()
    {
        return this.result;
    }

    @Override
    public PathPlanning setFrom(EntityID id)
    {
        this.from = id;
        return this;
    }

    @Override
    public PathPlanning setDestination(Collection<EntityID> targets)
    {
        this.targets = targets;
        return this;
    }

    @Override
    public PathPlanning precompute(PrecomputeData precomputeData)
    {
        super.precompute(precomputeData);
        return this;
    }

    @Override
    public PathPlanning resume(PrecomputeData precomputeData)
    {
        super.resume(precomputeData);
        return this;
    }

    @Override
    public PathPlanning preparate()
    {
        super.preparate();
        return this;
    }

    @Override
    public PathPlanning updateInfo(MessageManager messageManager)
    {
        return this;
    }

    @Override
    public PathPlanning calc()
    {
        ArrayList<EntityID> impassableList = new ArrayList<>();

        boolean found = false;
        boolean isSecond = false;
        do
        {
            found = false;
            List<EntityID> open = new LinkedList<>();
            Map<EntityID, EntityID> ancestors = new HashMap<>();
            open.add(this.from);
            EntityID next;
            ancestors.put(this.from, this.from);
            do
            {
                next = open.remove(0);
                if (isGoal(next, targets))
                {
                    found = true;
                    break;
                }
                Collection<EntityID> neighbours = graph.get(next);
                if (neighbours.isEmpty())
                {
                    continue;
                }
                for (EntityID neighbour : neighbours)
                {
                    if (isGoal(neighbour, targets))
                    {
                        ancestors.put(neighbour, next);
                        next = neighbour;
                        found = true;
                        break;
                    }
                    else
                    {
                        if (!impassableList.contains(neighbour) && !ancestors.containsKey(neighbour))
                        {
                            open.add(neighbour);
                            ancestors.put(neighbour, next);
                        }
                    }
                }
            } while (!found && !open.isEmpty());

            if (!found)
            {
                // No path
                isSecond = true;
                impassableList.clear();
                this.result = null;
            }
            // Walk back from goal to this.from
            EntityID current = next;
            LinkedList<EntityID> path = new LinkedList<>();
            do
            {
                path.add(0, current);
                current = ancestors.get(current);
                if (current == null)
                {
                    throw new RuntimeException("Found a node with no ancestor! Something is broken.");
                }
            } while (current != this.from);
            this.result = path;

            if (path.size() > 2)
            {
                for (int i = 1; i < (path.size() -1); i++)
                {
                    if (!checkAreaPassable.check((Area)worldInfo.getEntity(path.get(i)), (Area)worldInfo.getEntity(path.get(i -1)), (Area)worldInfo.getEntity(path.get(i +1)), worldInfo))
                    {
                        found = false;
                        impassableList.add(path.get(i));
                    }
                }
            }

        } while (!found);

        return this;
    }

    private boolean isGoal(EntityID e, Collection<EntityID> test)
    {
        return test.contains(e);
    }
}
