package naito.module.complex;

import adf.agent.communication.MessageManager;
import adf.agent.info.AgentInfo;
import adf.agent.info.ScenarioInfo;
import adf.agent.info.WorldInfo;
import adf.agent.module.ModuleManager;
import adf.agent.precompute.PrecomputeData;
import adf.component.module.complex.HumanSelector;
import rescuecore2.standard.entities.Human;
import rescuecore2.standard.entities.StandardEntityURN;
import rescuecore2.worldmodel.EntityID;

import java.util.*;

public class KKVictimSelector extends HumanSelector
{
    private EntityID result;
    private Set<Integer> knownVictims;


    public KKVictimSelector(
        AgentInfo     ai,
        WorldInfo     wi,
        ScenarioInfo  si,
        ModuleManager moduleManager)
    {
        super(ai, wi, si, moduleManager);
        this.knownVictims = new HashSet<>();
    }

    private boolean isDead(Human human)
    {
        return human.isHPDefined() && human.getHP() <= 0;
    }

    private boolean isHPSatisfied(Human human)
    {
        return !human.isHPDefined() || (human.getHP() > 3000);
    }

    private boolean isBuriedSatisfied(Human human)
    {
        return !human.isBuriednessDefined() || (human.getBuriedness() > 0);
    }

    private boolean isDamageSatisfied(Human human) {
        return !human.isDamageDefined() || (human.getDamage() > 0);
    }

    @Override
    public HumanSelector updateInfo(MessageManager messageManager)
    {
        Human victim;

        if (scenarioInfo.isDebugMode())
            System.err.println("( " + agentInfo.getTime() + ") " + agentInfo.getID() + " pre-knowledge : " + knownVictims);

        for (EntityID change : this.worldInfo.getChanged().getChangedEntities())
        {
            // myself
            if (change.equals(this.agentInfo.getID()))
                continue;

            if (change.equals(this.agentInfo.getPosition()) && this.knownVictims.contains(change))
            {
                // Area, and search
                this.knownVictims.remove(change.getValue());
                if (scenarioInfo.isDebugMode())
                    System.err.println("( " + agentInfo.getTime() + ") " + agentInfo.getID() + " Now reached Message Position : " + change.getValue());
            }

            // if human
            if (this.worldInfo.getEntity(change) instanceof Human)
            {
                victim = (Human) worldInfo.getEntity(change);

                // victim
                if (!this.knownVictims.contains(change.getValue()))
                {   // not registered (unknown agent or not needed help)
                    // maybe die
                    if (this.isDead(victim)) continue;

                    // or living, needs help ?
                    if (this.isBuriedSatisfied(victim))
                    {
                        this.knownVictims.add(change.getValue());
                        if (scenarioInfo.isDebugMode())
                            System.err.println("( " + agentInfo.getTime() + ") " + agentInfo.getID() + " Buried Agent (" + change.getValue() + " : " + (victim.isBuriednessDefined() ? victim.getBuriedness() : "null") + ")");
                    }
                }
                else
                {   // registered agents
                    if (this.worldInfo.getPosition(victim).getStandardURN() == StandardEntityURN.REFUGE ||
                        this.worldInfo.getPosition(victim).getStandardURN() == StandardEntityURN.AMBULANCE_TEAM
                    ){
                        // rescued
                        this.knownVictims.remove((Integer)change.getValue());
                        if (scenarioInfo.isDebugMode())
                            System.err.println("( " + agentInfo.getTime() + ") " + agentInfo.getID() + " Civilian Rescued (" + change.getValue() + ")");
                    }
                    else if(victim.getStandardURN() != StandardEntityURN.CIVILIAN
                            && !isBuriedSatisfied(victim)
                    ) {
                        // victim not civilian, and rescue ended
                        knownVictims.remove(change.getValue());
                        if (scenarioInfo.isDebugMode())
                            System.err.println("( " + agentInfo.getTime() + ") " + agentInfo.getID() + " Agent Rescued (" + change.getValue() + " : " + (victim.isBuriednessDefined() ? victim.getBuriedness() : "null") + ")");
                    }
                    else if (!isHPSatisfied(victim))
                    {
                        // maybe die
                        knownVictims.remove(change.getValue());
                        if (scenarioInfo.isDebugMode())
                            System.err.println("( " + agentInfo.getTime() + ") " + agentInfo.getID() + " Agent'll be die (" + change.getValue() + " : " + (victim.isHPDefined() ? victim.getHP() : "null") +  ")");
                    }
                }
            } //// fi entity is human
        }

        if (scenarioInfo.isDebugMode()) System.err.println("( " + agentInfo.getTime() + ") " + agentInfo.getID() + " post-knowledge : " + knownVictims);

        return this;
    }

    @Override
    public HumanSelector calc() {
        if(this.agentInfo.someoneOnBoard() != null) {
            this.result = this.agentInfo.someoneOnBoard().getID();
            return this;
        }

        Comparator<Integer> cmp = (o1, o2) -> worldInfo.getDistance(agentInfo.getPosition(), new EntityID(o1))
                - worldInfo.getDistance(agentInfo.getPosition(), new EntityID(o2));


        // search most nearest target
        this.result = knownVictims
                .stream()
                .min(cmp)
                .map(a -> new EntityID(a))
                .orElse(null);

        return this;
    }

    @Override
    public EntityID getTarget()
    {
        return this.result;
    }

    @Override
    public HumanSelector precompute(PrecomputeData precomputeData)
    {
        super.precompute(precomputeData);
        return this;
    }

    @Override
    public HumanSelector resume(PrecomputeData precomputeData)
    {
        super.resume(precomputeData);
        return this;
    }

    @Override
    public HumanSelector preparate() {
        super.preparate();
        return this;
    }
}
