package naito.module.complex;

import adf.agent.communication.MessageManager;
import adf.agent.communication.standard.bundle.information.MessageBuilding;
import adf.agent.info.AgentInfo;
import adf.agent.info.ScenarioInfo;
import adf.agent.info.WorldInfo;
import adf.agent.module.ModuleManager;
import adf.agent.precompute.PrecomputeData;
import adf.component.communication.CommunicationMessage;
import adf.component.module.algorithm.Clustering;
import adf.component.module.algorithm.PathPlanning;
import adf.component.module.complex.BuildingSelector;
import naito.SampleModuleKey;
import adf.util.WorldUtil;
import rescuecore2.standard.entities.Building;
import rescuecore2.standard.entities.FireBrigade;
import rescuecore2.standard.entities.StandardEntity;
import rescuecore2.worldmodel.EntityID;

import java.util.*;

import static rescuecore2.standard.entities.StandardEntityURN.*;

public class UBuildingSelector extends BuildingSelector {

    private int  AVERAGE_MOVEMENT_DISTANCE_PER_STEP = 40000;
    private static String KEY_FIRE_CLUSTERING_FOR_FIRE = "adf.sample.module.algorithm.UFireClustering";

    private int maxPower;
    private int maxDistance;
    private int refillRefuge;
    private int refillHydrant;

    private Map<EntityID, Integer> fireFightingDifficulty;

    private Set<Building> nearBuildings;

    private int updatedTime;

    private EntityID result;

    private int clusterIndex;

    public UBuildingSelector(AgentInfo ai, WorldInfo wi, ScenarioInfo si, ModuleManager moduleManager) {
        super(ai, wi, si, moduleManager);
        this.clusterIndex = -1;
        this.init();
    }

    private void init() {
        this.result = null;
        this.fireFightingDifficulty = new HashMap<>();
        this.maxDistance = this.scenarioInfo.getFireExtinguishMaxDistance();
        this.maxPower = this.scenarioInfo.getFireExtinguishMaxSum();
        this.refillRefuge = this.scenarioInfo.getFireTankRefillRate();
        this.refillHydrant = this.scenarioInfo.getFireTankRefillHydrantRate();
        this.nearBuildings = new HashSet<>();
        this.updatedTime = 0;
    }


    @Override
    public BuildingSelector calc() {
        return this.oldCalc2();
    }

    private BuildingSelector oldCalc() {
        Clustering clustering = this.moduleManager.getModule(SampleModuleKey.FIRE_MODULE_CLUSTERING);
        if(this.clusterIndex == -1) {
            this.clusterIndex = clustering.getClusterIndex(this.agentInfo.getID());
        }
        List<Building> buildingList = new ArrayList<>();
        for (StandardEntity next : clustering.getClusterEntities(this.clusterIndex)) {
            if (next.getStandardURN().equals(BUILDING)) {
                Building b = (Building)next;
                if (b.isOnFire()) {
                    buildingList.add(b);
                }
            }
        }
        // Sort by distance
        buildingList.sort(new DistanceSorter(this.worldInfo, this.agentInfo.getPositionArea()));
        this.result = buildingList.isEmpty() ? this.failedClusteringCalc() : buildingList.get(0).getID();
        return this;
    }

    private BuildingSelector oldCalc2() {
        Clustering clustering = this.moduleManager.getModule(SampleModuleKey.FIRE_MODULE_CLUSTERING);
        if(this.clusterIndex == -1) {
            this.clusterIndex = clustering.getClusterIndex(this.agentInfo.getID());
        }
        List<Building> buildingList = new ArrayList<>();
        for (StandardEntity next : clustering.getClusterEntities(this.clusterIndex)) {
            if (next.getStandardURN().equals(BUILDING)) {
                Building b = (Building)next;
                if (b.isOnFire()) {
                    buildingList.add(b);
                }
            }
        }
        // Sort by distance
        buildingList.sort(new DifficultySorter(this.fireFightingDifficulty, this.worldInfo, agentInfo.me()));
        this.result = buildingList.isEmpty() ? this.failedClusteringCalc() : buildingList.get(0).getID();
        return this;
    }

    private BuildingSelector newCalc() {
        this.result = null;
        List<StandardEntity> list = new ArrayList<>();
        Clustering clustering = this.moduleManager.getModule("", KEY_FIRE_CLUSTERING_FOR_FIRE);
        if(clustering != null) {
            for(Building building : this.nearBuildings) {
                this.result = this.selectTarget(clustering, building);
                if(this.result != null) {
                    System.out.println("<<<<<<<<<< use Fire clustering 1 >>>>>>>>>>");
                    return this;
                }
            }
            list.addAll(this.worldInfo.getEntitiesOfType(GAS_STATION));
            if(list.size() > 1) {
                list.sort(new DifficultySorter(this.fireFightingDifficulty, this.worldInfo, agentInfo.me()));
                for (StandardEntity entity : list) {
                    this.result = this.selectTarget(clustering, entity);
                    if (this.result != null) {
                        System.out.println("<<<<<<<<<< use Fire clustering 2 >>>>>>>>>>");
                        return this;
                    }
                }
            }
            list.clear();
            list.addAll(this.worldInfo.getEntitiesOfType(BUILDING, AMBULANCE_CENTRE, FIRE_STATION, POLICE_OFFICE));
            list.sort(new DifficultySorter(this.fireFightingDifficulty, this.worldInfo, agentInfo.me()));
            for(StandardEntity entity : list) {
                this.result = this.selectTarget(clustering, entity);
                if(this.result != null) {
                    System.out.println("<<<<<<<<<< use Fire clustering 3 >>>>>>>>>>");
                    return this;
                }
            }
        }
        return this.oldCalc();

    }

    private EntityID selectTarget(Clustering clustering, StandardEntity entity) {
        int index = clustering.getClusterIndex(entity);
        if(index == -1) {
            return null;
        }
        Collection<EntityID> elements = clustering.getClusterEntityIDs(index);
        if(elements == null || elements.isEmpty()) {
            return null;
        }
        Set<EntityID> cluster = new HashSet<>(elements);
        return this.selectTarget(cluster);
    }

    private EntityID selectTarget(Collection<EntityID> cluster) {
        List<Building> list = new ArrayList<>(cluster.size());
        StandardEntity me = this.agentInfo.me();
        for(EntityID id : cluster) {
            Building entity = (Building) this.worldInfo.getEntity(id);
            list.add(entity);
        }
        if(list.size() > 0) {
            list.sort(new DifficultySorter(this.fireFightingDifficulty, this.worldInfo, me));
            for(Building building : list) {
                if(building.isOnFire()) {
                    int fieryness = building.getFieryness();
                    if(fieryness == 1 || fieryness == 2) {
                        return building.getID();
                    }
                }
            }
        }
        return null;
    }

    private EntityID failedClusteringCalc() {
        List<Building> buildingList = new ArrayList<>();
        for (StandardEntity next : this.worldInfo.getEntitiesOfType(BUILDING)) {
            Building b = (Building)next;
            if (b.isOnFire()) {
                buildingList.add(b);
            }
        }
        // Sort by distance
        buildingList.sort(new DistanceSorter(this.worldInfo, this.agentInfo.getPositionArea()));
        return buildingList.isEmpty() ? null : buildingList.get(0).getID();
    }

    @Override
    public BuildingSelector updateInfo(MessageManager messageManager) {
        super.updateInfo(messageManager);
        // check updated
        int time = this.agentInfo.getTime();
        if(time == this.updatedTime) { return this; }
        // update
        this.updatedTime = time;
        Clustering clustering = moduleManager.getModule("", KEY_FIRE_CLUSTERING_FOR_FIRE);
        if(clustering == null) return this;
        clustering = clustering.updateInfo(messageManager);
        // changed
        this.updateByChanged(clustering);
        // message
        this.updateByMessage(clustering, messageManager);
        return  this;
    }



    private void updateByChanged(Clustering clustering ) {
        this.nearBuildings.clear();
        for(EntityID id : this.worldInfo.getChanged().getChangedEntities()) {
            StandardEntity entity = this.worldInfo.getEntity(id);
            if(entity instanceof Building) {
                this.nearBuildings.add((Building) entity);
            } /*else if(entity.getStandardURN() == FIRE_BRIGADE) {
                FireBrigade fb = (FireBrigade) entity;
                if(fb.isHPDefined() && fb.getHP() == 0) {

                } else if(fb.isWaterDefined() && (fb.getWater() < this.maxPower)) {

                } else {
                    boolean f = true;
                    for(StandardEntity entity1 : this.worldInfo.getObjectsInRange(fb, this.maxDistance)) {
                        if(entity1 instanceof Building && ((Building)entity1).isOnFire()) {
                            int index = clustering.getClusterIndex(entity1);
                            if(index != -1) {
                                this.fbTarget.put(id, entity1.getID());
                                this.resetAgent(id, index);
                                f = false;
                                break;
                            }
                        }
                    }
                    if(f) {
                        this.fbTarget.put(id, null);
                        this.resetAgent(id, ACTION_UNKNOWN);
                    }
                }
            }*/
        }
    }

    private void updateByMessage(Clustering clustering, MessageManager messageManager) {
        Set<EntityID> checked = new HashSet<>();
        for(CommunicationMessage message : messageManager.getReceivedMessageList()) {
            if(MessageBuilding.class.equals(message.getClass())) {
                MessageBuilding mb = (MessageBuilding)message;
                if(!this.nearBuildings.contains(this.worldInfo.getEntity(mb.getBuildingID()))) {
                    if(!checked.contains(mb.getBuildingID())) {
                        Building building = WorldUtil.reflectMessage(this.worldInfo, mb);
                        checked.add(building.getID());
                    }
                }
            }
        }
    }

    ////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

    @Override
    public EntityID getTarget() {
        return this.result;
    }

    @Override
    public BuildingSelector precompute(PrecomputeData precomputeData) {
        super.precompute(precomputeData);
        if(this.getCountPrecompute() > 0) { return this; }
        //values
        PathPlanning pathPlanning = moduleManager.getModule(SampleModuleKey.FIRE_MODULE_PATH_PLANNING, "adf.sample.module.algorithm.SamplePathPlanning");
        pathPlanning.precompute(precomputeData);
        List<EntityID> buildingKey = new ArrayList<>();
        Collection<StandardEntity> refillArea = this.worldInfo.getEntitiesOfType(REFUGE, HYDRANT);
        for(StandardEntity building : this.worldInfo.getEntitiesOfType(BUILDING, GAS_STATION, AMBULANCE_CENTRE, FIRE_STATION, POLICE_OFFICE)) {
            int bestCost = Integer.MAX_VALUE;
            EntityID buildingID = building.getID();
            for(StandardEntity area : refillArea) {
                int cost = this.getCost(buildingID, area, pathPlanning);
                if(bestCost > cost && cost > 0) { bestCost = cost; }
            }
            buildingKey.add(buildingID);
            fireFightingDifficulty.put(buildingID, bestCost);
        }
        for(EntityID id : buildingKey) {
            precomputeData.setInteger("fireFightingDifficulty-" + id.getValue(), this.fireFightingDifficulty.get(id));
        }
        return this;
    }

    @Override
    public BuildingSelector resume(PrecomputeData precomputeData) {
        super.resume(precomputeData);
        if(this.getCountResume() > 0) { return this; }

        PathPlanning pathPlanning = moduleManager.getModule(SampleModuleKey.FIRE_MODULE_PATH_PLANNING, "adf.sample.module.algorithm.SamplePathPlanning");
        pathPlanning.resume(precomputeData);
        for(StandardEntity entity : this.worldInfo.getEntitiesOfType(BUILDING, GAS_STATION, AMBULANCE_CENTRE, FIRE_STATION, POLICE_OFFICE)) {
            EntityID buildingID = entity.getID();
            this.fireFightingDifficulty.put(buildingID, precomputeData.getInteger("fireFightingDifficulty-" + buildingID.getValue()));
        }
        return this;
    }

    @Override
    public BuildingSelector preparate() {
        super.preparate();
        if(this.getCountPreparate() > 0) { return this; }

        Collection<StandardEntity> refillArea = this.worldInfo.getEntitiesOfType(REFUGE, HYDRANT);
        for(StandardEntity building : this.worldInfo.getEntitiesOfType(BUILDING, GAS_STATION, AMBULANCE_CENTRE, FIRE_STATION, POLICE_OFFICE)) {
            int bestCost = Integer.MAX_VALUE;
            for(StandardEntity area : refillArea) {
                int cost = this.getSimpleCost(building, area);
                if(bestCost > cost && cost > 0) { bestCost = cost; }
            }
            fireFightingDifficulty.put(building.getID(), bestCost);
        }
        return this;
    }

    private int getSimpleCost(StandardEntity building, StandardEntity refillArea) {
        int cost;
        if(refillArea.getStandardURN() == REFUGE) {
            cost = this.maxPower / this.refillRefuge;
            if((this.maxPower % this.refillRefuge) > 0) { cost++; }
        } else {
            cost = this.maxPower / this.refillHydrant;
            if((this.maxPower % this.refillHydrant) > 0) { cost++; }
        }
        double d = this.worldInfo.getDistance(refillArea, building) - this.maxDistance;
        int dCost = (int)(d / AVERAGE_MOVEMENT_DISTANCE_PER_STEP);
        if((d % AVERAGE_MOVEMENT_DISTANCE_PER_STEP) > 0) { dCost++; }
        if(d < 0) { dCost = 0; }
        // int FIRE_SPEED = 5; cost += (dCost / FIRE_SPEED);
        return cost + dCost;
    }

    private int getCost(EntityID buildingID, StandardEntity refillArea, PathPlanning pathPlanning) {
        int cost;
        if(refillArea.getStandardURN() == REFUGE) {
            cost = this.maxPower / this.refillRefuge;
            if((this.maxPower % this.refillRefuge) > 0) { cost++; }
        } else {
            cost = this.maxPower / this.refillHydrant;
            if((this.maxPower % this.refillHydrant) > 0) { cost++; }
        }
        double d = pathPlanning.getDistance(refillArea.getID(), buildingID) - this.maxDistance;
        int dCost = (int)(d / AVERAGE_MOVEMENT_DISTANCE_PER_STEP);
        if((d % AVERAGE_MOVEMENT_DISTANCE_PER_STEP) > 0) { dCost++; }
        if(d < 0) { dCost = 0; }
        // int FIRE_SPEED = 5; cost += (dCost / FIRE_SPEED);
        return cost + dCost;
    }

    private class DifficultySorter implements Comparator<StandardEntity> {
        private Map<EntityID, Integer> difficulty;
        private StandardEntity reference;
        private WorldInfo worldInfo;

        private int AVERAGE_MOVEMENT_DISTANCE_PER_STEP = 40000;

        DifficultySorter(Map<EntityID, Integer> difficulty, WorldInfo wi, StandardEntity reference) {
            this.reference = reference;
            this.worldInfo = wi;
            this.difficulty = difficulty;
        }

        @Override
        public int compare(StandardEntity o1, StandardEntity o2) {
            Integer c1 = this.difficulty.get(o1.getID());
            Integer c2 = this.difficulty.get(o2.getID());
            if(c1 == null) {
                return c2 == null ? 0 : -1;
            }
            if(c2 == null) return 1;
            int d1 = this.worldInfo.getDistance(this.reference, o1) / AVERAGE_MOVEMENT_DISTANCE_PER_STEP;
            int d2 = this.worldInfo.getDistance(this.reference, o2) / AVERAGE_MOVEMENT_DISTANCE_PER_STEP;

            return (c1 + d1) - (c2 + d2);
        }
    }

    private class DistanceSorter implements Comparator<StandardEntity> {
        private StandardEntity reference;
        private WorldInfo worldInfo;

        DistanceSorter(WorldInfo wi, StandardEntity reference) {
            this.reference = reference;
            this.worldInfo = wi;
        }

        public int compare(StandardEntity a, StandardEntity b) {
            int d1 = this.worldInfo.getDistance(this.reference, a);
            int d2 = this.worldInfo.getDistance(this.reference, b);
            return d1 - d2;
        }
    }
}
