package naito.module.complex.mroadselector;

import adf.agent.info.AgentInfo;
import adf.agent.info.WorldInfo;
import adf.agent.info.ScenarioInfo;
import adf.component.module.algorithm.PathPlanning;
import adf.agent.module.ModuleManager;
import adf.agent.precompute.PrecomputeData;
import adf.component.module.algorithm.Clustering;
import adf.component.extaction.ExtAction;
import adf.component.module.complex.RoadSelector;
import adf.component.communication.CommunicationMessage;
import adf.agent.communication.MessageManager;
import adf.agent.communication.standard.bundle.information.MessageAmbulanceTeam;
import adf.agent.communication.standard.bundle.information.MessageFireBrigade;
import adf.agent.communication.standard.bundle.information.MessagePoliceForce;
import adf.agent.action.common.ActionRest;
import naito.SampleModuleKey;
import rescuecore2.standard.entities.*;
import rescuecore2.worldmodel.EntityID;
import rescuecore2.misc.geometry.GeometryTools2D;
import rescuecore2.misc.geometry.Point2D;
import rescuecore2.misc.geometry.Line2D;

import java.util.*;

public class MRoadSelector extends RoadSelector
{
    private EntityID                 result;
    private Map<EntityID,TargetType> candidateTypeMap;
    private Set<EntityID> candidatePenalty;
    private EntityID goalResult;

    private Set<EntityID> preventTurnRequesters;

    private List<EntityID> path;
    private Integer        pathIndex;

    private static final double AGENT_SIZE = 500;

    public MRoadSelector(
        AgentInfo     ai,
        WorldInfo     wi,
        ScenarioInfo  si,
        ModuleManager mm
    ){
        super(ai, wi, si, mm);

        this.result = null;
        this.candidateTypeMap = new HashMap<>();
        this.preventTurnRequesters = new HashSet<>();
        this.candidatePenalty = new HashSet<>();
    }

    @Override
    public RoadSelector calc()
    {
        if (this.agentInfo.getTime() <= scenarioInfo.getKernelAgentsIgnoreuntil())
        {
            this.result = null;
            return this;
        }

        if (this.result != null && this.hasDoneClear(this.goalResult))
        {
            this.candidateTypeMap.remove(this.goalResult);
            this.goalResult = null;
        }

        List<EntityID> candidateIDs
            = new ArrayList<>(this.candidateTypeMap.keySet());
        Map<EntityID,Integer> candidateImportanceMap
            = this.createImportanceMap(this.candidateTypeMap);

        candidateIDs.sort( (eid1, eid2) ->
            candidateImportanceMap.get(eid1).compareTo(
                candidateImportanceMap.get(eid2)
            )
        );
        //Collections.reverse(candidateIDs);

        if (candidateIDs.size() == 0)
        {
            this.result = null;
            this.goalResult = null;
            System.out.println(this.getClass().getSimpleName() + ": result=" + this.result);
            return this;
        }

        EntityID highestImportanceGoalID = candidateIDs.get(0);
        if (goalResult == null || !candidateIDs.contains(goalResult))
        {
            this.goalResult = highestImportanceGoalID;
            this.result = this.decideTarget(highestImportanceGoalID);
            System.out.println(this.getClass().getSimpleName() + ": result=" + this.result);
            return this;
        }

        int presentGoalImportance = candidateImportanceMap.get(this.goalResult);
        int highestImportance = candidateImportanceMap.get(highestImportanceGoalID);

        if (highestImportance < presentGoalImportance)
        {
            this.goalResult = highestImportanceGoalID;
            this.result = this.decideTarget(highestImportanceGoalID);
        }
        else
        {
            this.result = this.decideTarget(this.goalResult);
        }

        System.out.println(this.getClass().getSimpleName() + ": result=" + this.result);
        return this;
    }

    private EntityID decideTarget(EntityID goal)
    {
        if (!goal.equals(Util.getLast(this.path)) || !this.path.contains(this.agentInfo.getPosition()))
        {
            this.path = this.calculatePath(goal);
            this.pathIndex = 0;
        }

        if (this.path == null)
        {
            if (this.agentInfo.getPosition().equals(goal))
                this.candidateTypeMap.remove(goal);

            return null;
        }

        if (this.path.size() == 0)
        {
            if (this.agentInfo.getPosition().equals(goal))
            {
                this.candidateTypeMap.remove(goal);
            }
            return null;
        }

        this.pathIndex = this.findNextClearArea(this.path, this.pathIndex);
        EntityID trueTarget = this.path.get(this.pathIndex);

        if (trueTarget == null)
        {
            if (this.agentInfo.getPosition().equals(goal))
                this.candidateTypeMap.remove(goal);

            return null;
        }

        StandardEntity standardEntity = this.worldInfo.getEntity(trueTarget);
        if (!(standardEntity instanceof Area))
        {
            if (this.agentInfo.getPosition().equals(goal))
                this.candidateTypeMap.remove(goal);

            return null;
        }

        Area targetArea = (Area)standardEntity;
        if (!targetArea.isBlockadesDefined())
        {
            if (trueTarget.equals(goal))
                this.candidateTypeMap.remove(goal);

            return null;
        }

        if (targetArea.getBlockades().isEmpty() && trueTarget.equals(goal))
        {
            this.candidateTypeMap.remove(goal);
            return null;
        }

        double bestDistance = Double.MAX_VALUE;
        Blockade bestBlockade = null;

        for (EntityID entityID : targetArea.getBlockades())
        {
            Blockade blockade = (Blockade)this.worldInfo.getEntity(entityID);
            List<Line2D> lines =
                GeometryTools2D.pointsToLines(GeometryTools2D.vertexArrayToPoints(blockade.getApexes()), true);
            Point2D origin = new Point2D(this.agentInfo.getX(), this.agentInfo.getY());

            for (Line2D next : lines)
            {
                Point2D closest = GeometryTools2D.getClosestPointOnSegment(next, origin);
                double d = GeometryTools2D.getDistance(origin, closest);
                if (d < bestDistance) {
                    bestDistance = d;
                    bestBlockade = blockade;
                }
            }
        }

        if (bestBlockade == null)
            return trueTarget;

        if (bestDistance < this.scenarioInfo.getClearRepairDistance()/2)
            return bestBlockade.getID();

        return trueTarget;
    }

    private List<EntityID> calculatePath(EntityID destination)
    {
        PathPlanning pathPlanning =
            this.moduleManager.getModule(SampleModuleKey.POLICE_MODULE_PATH_PLANNING);

        pathPlanning.setFrom(this.agentInfo.getPosition()).setDestination(destination);
        List<EntityID> path = pathPlanning.calc().getResult();
        path.add(0, this.agentInfo.getPosition());

        return path;
    }

    private int findNextClearArea(List<EntityID> path, int index)
    {
        Clustering areaIsUpdatedInfo = this.moduleManager.getModule(
            "",
            "naito.module.algorithm.MAreaIsUpdatedInfo"
        );
        Set<EntityID> areaIsUpdatedInfoSet = new HashSet<>(
            areaIsUpdatedInfo.getClusterEntityIDs(0)
        );

        System.out.println(areaIsUpdatedInfo.getClass());

        for (; index<path.size(); index++)
        {
            EntityID entityID = path.get(index);
            StandardEntity entity = this.worldInfo.getEntity(entityID);

            if (index == path.size()-1)
                break;
            if (entity instanceof Building)
                continue;
            if (!(entity instanceof Area))
                break;
            if (!areaIsUpdatedInfoSet.contains(entityID))
                break;

            //if (this.findHumanInBlockade(entityID).size() != 0)
            //    break;

            Area area = (Area)entity;
            if (area.isBlockadesDefined() && !area.getBlockades().isEmpty()) break;
        }

        return index;
    }

    private boolean hasDoneClear(EntityID areaID)
    {
        Clustering areaIsUpdatedInfo = this.moduleManager.getModule(
            "",
            "naito.module.algorithm.MAreaIsUpdatedInfo"
        );
        Set<EntityID> areaIsUpdatedInfoSet = new HashSet<>(
            areaIsUpdatedInfo.getClusterEntityIDs(0)
        );
        StandardEntity entity = this.worldInfo.getEntity(areaID);
        if (!(entity instanceof Area))
            return true;
        if (!areaIsUpdatedInfoSet.contains(areaID))
            return false;

        Area area = (Area)entity;
        if (area.isBlockadesDefined() && !area.getBlockades().isEmpty())
            return false;

        return true;
    }

    private Map<EntityID,Integer> createImportanceMap(
        Map<EntityID,TargetType> candidateTypeMap
    ){
        Map<EntityID,Integer> importanceMap = new HashMap<>();
        List<EntityID> candidateIDs = new ArrayList<>(candidateTypeMap.keySet());

        for (EntityID entityID : candidateIDs)
        {
            Area area = (Area)this.worldInfo.getEntity(entityID);

            importanceMap.put(
                entityID,
                this.calculateImportance(area, candidateTypeMap.get(entityID))
            );
        }

        return importanceMap;
    }

    private static int AVERAGE_MOV_DISTANCE = 40000;

    private int calculateImportance(Area area, TargetType targetType)
    {
        double distance = GeometryTools2D.getDistance(
            Util.toPoint2D(this.agentInfo),
            Util.toPoint2D(area)
        );

        //int stepToMove = (int)(distance / AVERAGE_MOV_DISTANCE * 1.5);
        int clearRepairDistance = this.scenarioInfo.getClearRepairDistance();
        int stepToMove = (int)(distance / clearRepairDistance * 1.2);

        int penalty = (this.candidatePenalty.contains(area.getID())) ? 3 : 1;

        return (targetType.getBaseImportance() + stepToMove) * penalty;
    }

    @Override
    public EntityID getTarget()
    {
        return this.result;
    }

    @Override
    public RoadSelector precompute(PrecomputeData precomputeData)
    {
        super.precompute(precomputeData);
        Clustering areaIsUpdatedInfo = moduleManager.getModule(
            "",
            "naito.module.algorithm.MAreaIsUpdatedInfo"
        );
        areaIsUpdatedInfo.precompute(precomputeData);
        return this;
    }

    @Override
    public RoadSelector resume(PrecomputeData precomputeData)
    {
        super.resume(precomputeData);

        Clustering areaIsUpdatedInfo = moduleManager.getModule(
            "",
            "naito.module.algorithm.MAreaIsUpdatedInfo"
        );
        areaIsUpdatedInfo.resume(precomputeData);
        this.addRefugesToCandidates();
        this.addHydrantsToCandidates();

        return this;
    }

    @Override
    public RoadSelector preparate()
    {
        super.preparate();

        this.addRefugesToCandidates();
        this.addHydrantsToCandidates();
        Clustering areaIsUpdatedInfo = moduleManager.getModule(
            "",
            "naito.module.algorithm.MAreaIsUpdatedInfo"
        );
        areaIsUpdatedInfo.preparate();

        return this;
    }

    private void addRefugesToCandidates()
    {
        List<StandardEntity> refuges = new ArrayList<>(
            this.worldInfo.getEntitiesOfType(StandardEntityURN.REFUGE)
        );

        for (StandardEntity entity : refuges)
            this.candidateTypeMap.put(entity.getID(), TargetType.FACILITY_REFUGE);

        return;
    }

    private void addHydrantsToCandidates()
    {
        List<StandardEntity> hydrants = new ArrayList<>(
            this.worldInfo.getEntitiesOfType(StandardEntityURN.HYDRANT)
        );

        for (StandardEntity entity : hydrants)
            this.candidateTypeMap.put(entity.getID(), TargetType.FACILITY_HYDRANT);

        return;
    }

    @Override
    public RoadSelector updateInfo(MessageManager messageManager)
    {
        super.updateInfo(messageManager);
        Clustering areaIsUpdatedInfo = moduleManager.getModule(
            "",
            "naito.module.algorithm.MAreaIsUpdatedInfo"
        );
        areaIsUpdatedInfo.updateInfo(messageManager);

        Set<EntityID> presentTurnRequesters = new HashSet<>();

        for (CommunicationMessage message : messageManager.getReceivedMessageList())
        {
            if (!message.isRadio() &&
                message instanceof MessagePoliceForce
            ){
                MessagePoliceForce messagePoliceForce = (MessagePoliceForce)message;

                if (messagePoliceForce.getAction() != MessagePoliceForce.ACTION_CLEAR)
                    continue;

                if (this.agentInfo.getID().getValue() < messagePoliceForce.getSenderID().getValue())
                    continue;

                EntityID targetID = messagePoliceForce.getTargetID();
                this.candidateTypeMap.put(targetID, TargetType.OTHER_AGENT);

                if (targetID == null || !targetID.equals(this.result))
                    continue;

                StandardEntity targetEntity = this.worldInfo.getEntity(targetID);
                if (!(targetEntity instanceof Area))
                    continue;

                Area targetArea = (Area)targetEntity;

                List<EntityID> candidateIDs
                    = new ArrayList<>(this.candidateTypeMap.keySet());

                for (EntityID candidateID : candidateIDs)
                {
                    StandardEntity candidateEntity = this.worldInfo.getEntity(candidateID);
                    if (!(candidateEntity instanceof Area))
                        continue;

                    Area candidate = (Area)candidateEntity;
                    double distance = GeometryTools2D.getDistance(
                        new Point2D(targetArea.getX(), targetArea.getY()),
                        new Point2D(candidate.getX(), candidate.getY())
                    );

                    if (distance < 20000)
                        this.candidatePenalty.add(candidateID);
                }
            }

            if (message instanceof MessageFireBrigade)
            {
                MessageFireBrigade messageFireBrigade = (MessageFireBrigade)message;
                if (messageFireBrigade.getAction() != MessageFireBrigade.ACTION_REST)
                    continue;

                Area position = (Area)this.worldInfo.getEntity(messageFireBrigade.getPosition());
                if (position.getStandardURN() == StandardEntityURN.REFUGE)
                    continue;

                EntityID targetID = messageFireBrigade.getTargetID();
                StandardEntity target = this.worldInfo.getEntity(targetID);
                if (!(target instanceof Area))
                    continue;

                EntityID senderID = messageFireBrigade.getSenderID();
                TargetType targetType = (this.preventTurnRequesters.contains(senderID))
                    ?   TargetType.OBSTACLE_ESCAPE
                    :   TargetType.OBSTACLE_TRAFFIC;


                if (this.candidateTypeMap.containsKey(targetID) &&
                    this.candidateTypeMap.get(targetID).isMoreImport(targetType)
                ){
                    continue;
                }

                if (this.candidateTypeMap.containsKey(targetID) &&
                    this.candidateTypeMap.get(targetID) == TargetType.OTHER_AGENT
                ){
                    continue;
                }

                this.candidateTypeMap.put(targetID, targetType);
                presentTurnRequesters.add(senderID);
            }

            if (message instanceof MessageAmbulanceTeam)
            {
                MessageAmbulanceTeam messageAmbulanceTeam = (MessageAmbulanceTeam)message;
                if (messageAmbulanceTeam.getAction() != MessageAmbulanceTeam.ACTION_REST)
                    continue;

                Area position = (Area)this.worldInfo.getEntity(messageAmbulanceTeam.getPosition());
                if (position.getStandardURN() == StandardEntityURN.REFUGE)
                    continue;

                EntityID targetID = messageAmbulanceTeam.getTargetID();
                StandardEntity target = this.worldInfo.getEntity(targetID);
                if (!(target instanceof Area))
                    continue;

                EntityID senderID = messageAmbulanceTeam.getSenderID();
                TargetType targetType = (this.preventTurnRequesters.contains(senderID))
                    ?   TargetType.OBSTACLE_ESCAPE
                    :   TargetType.OBSTACLE_TRAFFIC;

                if (this.candidateTypeMap.containsKey(targetID) &&
                    this.candidateTypeMap.get(targetID).isMoreImport(targetType)
                ){
                    continue;
                }

                if (this.candidateTypeMap.containsKey(targetID) &&
                    this.candidateTypeMap.get(targetID) == TargetType.OTHER_AGENT
                ){
                    continue;
                }

                this.candidateTypeMap.put(targetID, TargetType.OBSTACLE_TRAFFIC);
                presentTurnRequesters.add(senderID);
            }
        }

        this.preventTurnRequesters = presentTurnRequesters;

        Set<EntityID> areasExistHumanInBlockade = new HashSet<>();

        for (EntityID entityID : this.worldInfo.getChanged().getChangedEntities())
        {
            if (entityID.equals(this.agentInfo.getID()))
                continue;

            StandardEntity entity = this.worldInfo.getEntity(entityID);
            if (!(entity instanceof Human))
                continue;

            Human human = (Human)entity;
            EntityID positionID = human.getPosition();
            StandardEntity positionEntity = this.worldInfo.getEntity(positionID);

            if (!(positionEntity instanceof Area))
                continue;

            Area positionArea = (Area)positionEntity;
            if (!positionArea.isBlockadesDefined())
                continue;

            List<java.awt.Shape> blockadeShapes =
                Util.toShapes(Util.toBlockades(positionArea, this.worldInfo));

            for (java.awt.Shape blockadeShape : blockadeShapes)
            {
                if (blockadeShape.contains(human.getX(), human.getY()))
                {
                    areasExistHumanInBlockade.add(positionID);
                    this.candidateTypeMap.put(positionID, TargetType.HUMAN_IN_BLOCKADE);
                    break;
                }
            }
        }

        for (EntityID entityID : this.worldInfo.getChanged().getChangedEntities())
        {
            StandardEntity entity = this.worldInfo.getEntity(entityID);
            if (!(entity instanceof Area))
                continue;

            if (!this.candidateTypeMap.containsKey(entityID))
                continue;

            if (this.candidateTypeMap.get(entityID) != TargetType.HUMAN_IN_BLOCKADE)
                continue;

            if (areasExistHumanInBlockade.contains(entityID))
                continue;

            this.candidateTypeMap.remove(entityID);
        }

        return this;
    }
}
