package naito.tactics.mpolice;

import adf.agent.action.Action;
import adf.agent.action.common.ActionMove;
import adf.agent.action.common.ActionRest;
import adf.agent.action.police.ActionClear;
import adf.agent.communication.MessageManager;
import adf.agent.communication.standard.bundle.information.*;
import adf.agent.info.AgentInfo;
import adf.agent.info.ScenarioInfo;
import adf.agent.info.WorldInfo;
import adf.agent.module.ModuleManager;
import adf.agent.precompute.PrecomputeData;
import adf.component.communication.CommunicationMessage;
import adf.component.module.algorithm.Clustering;
import adf.component.module.algorithm.PathPlanning; import adf.component.module.complex.RoadSelector;
import adf.component.module.complex.Search;
import adf.component.tactics.TacticsPolice;
import adf.component.extaction.ExtAction;
import naito.SampleModuleKey;
import adf.util.WorldUtil;
import rescuecore2.standard.entities.*;
import rescuecore2.worldmodel.EntityID;
import rescuecore2.misc.geometry.*;

import java.util.*;

public class MPolice extends TacticsPolice
{
    private PathPlanning pathPlanning;
    private RoadSelector roadSelector;
    private Search       search;
    private Clustering   clustering;
    private Clustering   areaIsUpdatedInfo;

    private ActionState   actionState;
    private List<Point2D> positionHistory;
    private final static int AGENT_SIZE = 500;

    @Override
    public void initialize(
        AgentInfo      agentInfo,
        WorldInfo      worldInfo,
        ScenarioInfo   scenarioInfo,
        ModuleManager  moduleManager,
        MessageManager messageManager
    ){
        worldInfo.indexClass(
            StandardEntityURN.ROAD,
            StandardEntityURN.HYDRANT,
            StandardEntityURN.BUILDING,
            StandardEntityURN.REFUGE,
            StandardEntityURN.BLOCKADE
        );

        //init ExtAction
        moduleManager.getExtAction(
            SampleModuleKey.POLICE_ACTION_EXT_CLEAR,
            "naito.extaction.ActionExtClear"
        );
        moduleManager.getExtAction(
            SampleModuleKey.POLICE_ACTION_SEARCH,
            "naito.extaction.ActionSearch"
        );

        this.positionHistory = new LinkedList<>();
    }

    @Override
    public void precompute(
        AgentInfo      agentInfo,
        WorldInfo      worldInfo,
        ScenarioInfo   scenarioInfo,
        ModuleManager  moduleManager,
        PrecomputeData precomputeData
    ){
        this.pathPlanning = moduleManager.getModule(
            SampleModuleKey.POLICE_MODULE_PATH_PLANNING,
            "naito.module.algorithm.SamplePathPlanning"
        );
        this.pathPlanning.precompute(precomputeData);

        this.clustering = moduleManager.getModule(
            SampleModuleKey.POLICE_MODULE_CLUSTERING,
            "naito.module.algorithm.SampleKMeans"
        );
        this.clustering.precompute(precomputeData);

        this.search = moduleManager.getModule(
            SampleModuleKey.POLICE_MODULE_SEARCH,
            "naito.module.complex.SampleSearch"
        );
        this.search.precompute(precomputeData);

        this.roadSelector = moduleManager.getModule(
            SampleModuleKey.POLICE_MODULE_ROAD_SELECTOR,
            "naito.module.complex.SampleRoadSelector"
        );
        this.roadSelector.precompute(precomputeData);

        this.areaIsUpdatedInfo = moduleManager.getModule(
            "",
            "naito.module.algorithm.MAreaIsUpdatedInfo"
        );
        this.areaIsUpdatedInfo.precompute(precomputeData);
    }

    @Override
    public void resume(
        AgentInfo      agentInfo,
        WorldInfo      worldInfo,
        ScenarioInfo   scenarioInfo,
        ModuleManager  moduleManager,
        PrecomputeData precomputeData
    ){
        this.pathPlanning = moduleManager.getModule(
            SampleModuleKey.POLICE_MODULE_PATH_PLANNING,
            "naito.module.algorithm.SamplePathPlanning"
        );
        this.pathPlanning.resume(precomputeData);

        this.clustering = moduleManager.getModule(
            SampleModuleKey.POLICE_MODULE_CLUSTERING,
            "naito.module.algorithm.SampleKMeans"
        );
        this.clustering.resume(precomputeData);

        this.search = moduleManager.getModule(
            SampleModuleKey.POLICE_MODULE_SEARCH,
            "naito.module.complex.SampleSearch"
        );
        this.search.resume(precomputeData);

        this.roadSelector = moduleManager.getModule(
            SampleModuleKey.POLICE_MODULE_ROAD_SELECTOR,
            "naito.module.complex.SampleRoadSelector"
        );
        this.roadSelector.resume(precomputeData);

        this.areaIsUpdatedInfo = moduleManager.getModule(
            "",
            "naito.module.algorithm.MAreaIsUpdatedInfo"
        );
        this.areaIsUpdatedInfo.resume(precomputeData);
    }

    @Override
    public void preparate(
        AgentInfo     agentInfo,
        WorldInfo     worldInfo,
        ScenarioInfo  scenarioInfo,
        ModuleManager moduleManager
    ){
        this.pathPlanning = moduleManager.getModule(
            SampleModuleKey.POLICE_MODULE_PATH_PLANNING,
            "naito.module.algorithm.SamplePathPlanning"
        );
        this.pathPlanning.preparate();

        this.clustering = moduleManager.getModule(
            SampleModuleKey.POLICE_MODULE_CLUSTERING,
            "naito.module.algorithm.SampleKMeans"
        );
        this.clustering.preparate();

        this.search = moduleManager.getModule(
            SampleModuleKey.POLICE_MODULE_SEARCH,
            "naito.module.complex.SampleSearch"
        );
        this.search.preparate();

        this.roadSelector = moduleManager.getModule(
            SampleModuleKey.POLICE_MODULE_ROAD_SELECTOR,
            "naito.module.complex.SampleRoadSelector"
        );
        this.roadSelector.preparate();

        this.areaIsUpdatedInfo = moduleManager.getModule(
            "",
            "naito.module.algorithm.MAreaIsUpdatedInfo"
        );
        this.areaIsUpdatedInfo.preparate();
    }

    @Override
    public Action think(
        AgentInfo      agentInfo,
        WorldInfo      worldInfo,
        ScenarioInfo   scenarioInfo,
        ModuleManager  moduleManager,
        MessageManager messageManager
    ){
        this.pathPlanning.updateInfo(messageManager);
        this.clustering.updateInfo(messageManager);
        this.roadSelector.updateInfo(messageManager);
        this.search.updateInfo(messageManager);
        this.areaIsUpdatedInfo.updateInfo(messageManager);

        this.updateInfo(worldInfo, messageManager);
        this.sendMessage(worldInfo, messageManager);

        this.positionHistory.add(Util.toPoint2D(agentInfo));
        if (this.positionHistory.size() > 3)
            this.positionHistory.remove(0);

        PoliceForce me = (PoliceForce)agentInfo.me();
        //check buriedness
        if (me.isBuriednessDefined() && me.getBuriedness() > 0)
        {
            this.updateInfo(worldInfo, messageManager);
            messageManager.addMessage(
                new MessagePoliceForce(true, me, MessagePoliceForce.ACTION_REST, me.getPosition())
            );
            this.actionState = ActionState.Normal;
            return new ActionRest();
        }

        if (agentInfo.getTime() <= scenarioInfo.getKernelAgentsIgnoreuntil())
        {
            this.actionState = ActionState.Normal;
            return new ActionRest();
        }

        EntityID clearTarget = this.roadSelector.calc().getTarget();
        Action action;

        if (clearTarget != null)
        {
            ExtAction extActionClear = moduleManager.getExtAction(
                SampleModuleKey.POLICE_ACTION_EXT_CLEAR,
                "naito.extactoin.ActionExtClear"
            );
            messageManager.addMessage(
                new MessagePoliceForce(false, (PoliceForce)me, MessagePoliceForce.ACTION_CLEAR, clearTarget)
            );

            action = extActionClear.setTarget(clearTarget).calc().getAction();
        }
        else
        {
            EntityID searchTarget = this.search.calc().getTarget();

            ExtAction extActionSearch = moduleManager.getExtAction(
                SampleModuleKey.POLICE_ACTION_SEARCH,
                "naito.extaction.ActionSearch"
            );

            action = extActionSearch.setTarget(searchTarget).calc().getAction();
        }

        if (this.actionState == ActionState.InterruptClear)
        {
            this.actionState = ActionState.AfterInterruptClear;
            return action;
        }

        if (!this.isAgentStop())
        {
            this.actionState = ActionState.Normal;
            return action;
        }

        if (this.actionState == ActionState.Normal)
        {
            System.out.println("MPolice: InterruptClear");
            this.actionState = ActionState.InterruptClear;

            ExtAction extActionClear = moduleManager.getExtAction(
                SampleModuleKey.POLICE_ACTION_EXT_CLEAR,
                "naito.extactoin.ActionExtClear"
            );

            return extActionClear.setTarget(agentInfo.getPosition()).calc().getAction();
        }

        if (this.actionState == ActionState.AfterInterruptClear)
        {
            System.out.println("---------------------MPolice: InterruptMove--------------------");
            this.actionState = ActionState.InterruptMove;
            ExtAction extActionClear = moduleManager.getExtAction(
                SampleModuleKey.POLICE_ACTION_EXT_CLEAR,
                "naito.extactoin.ActionExtClear"
            );
            return extActionClear.setTarget(agentInfo.getPosition()).calc().getAction();
        }

        this.actionState = ActionState.Fail;
        return action;
    }

    private boolean isAgentStop()
    {
        if (this.positionHistory.size() < 3)
            return false;

        Point2D pos1 = this.positionHistory.get(this.positionHistory.size()-1);
        Point2D pos2 = this.positionHistory.get(this.positionHistory.size()-2);
        Point2D pos3 = this.positionHistory.get(this.positionHistory.size()-3);
        if (GeometryTools2D.getDistance(pos1, pos2) > AGENT_SIZE * 2)
            return false;

        if (GeometryTools2D.getDistance(pos2, pos3) > AGENT_SIZE * 2)
            return false;

        return true;
    }

    private EntityID getEntity(Point2D point, WorldInfo wi)
    {
        List<EntityID> entities = new ArrayList<>(
            wi.getObjectIDsInRange((int)point.getX(), (int)point.getY(), AGENT_SIZE)
        );

        if (entities.size() > 0)
            return entities.get(0);

        return null;
    }


    private void sendMessage(WorldInfo worldInfo, MessageManager messageManager)
    {
        for(EntityID id : worldInfo.getChanged().getChangedEntities())
        {
            StandardEntity entity = worldInfo.getEntity(id);

            if (entity.getStandardURN() == StandardEntityURN.CIVILIAN)
            {
                Civilian civilian = (Civilian)entity;
                if (this.needSend(worldInfo, civilian))
                    messageManager.addMessage(new MessageCivilian(true, civilian));
            }

            if (entity instanceof Building)
            {
                Building building = (Building)entity;

                if (!building.isFierynessDefined()) continue;

                if (building.getFieryness() >= 1 && building.getFieryness() <= 3)
                    messageManager.addMessage(new MessageBuilding(true, building));
            }

            if (entity instanceof Road)
            {
                Road road = (Road)entity;

                if (road.isBlockadesDefined() && road.getBlockades().isEmpty())
                    messageManager.addMessage(new MessageRoad(true, road, null, true));
            }
        }
    }

    private boolean needSend(WorldInfo wi, Civilian civilian)
    {
        if (civilian.isBuriednessDefined() && civilian.getBuriedness() > 0)
            return true;

        if (civilian.isDamageDefined()   &&
           civilian.getDamage() > 0     &&
           civilian.isPositionDefined() &&
           wi.getEntity(civilian.getPosition()) instanceof Road
        ){
            return true;
        }

        return false;
    }

    private void updateInfo(WorldInfo wi, MessageManager messageManager)
    {
        for (CommunicationMessage message : messageManager.getReceivedMessageList())
        {
            if (message.getClass() == MessageCivilian.class)
            {
                MessageCivilian mc = (MessageCivilian)message;
                if (!wi.getChanged().getChangedEntities().contains(mc.getAgentID()))
                    WorldUtil.reflectMessage(wi, mc);
            }

            if (message.getClass() == MessageBuilding.class)
            {
                MessageBuilding mb = (MessageBuilding)message;
                if (!wi.getChanged().getChangedEntities().contains(mb.getBuildingID()))
                    WorldUtil.reflectMessage(wi, mb);
            }

            if (message.getClass() == MessageAmbulanceTeam.class)
            {
                MessageAmbulanceTeam mat = (MessageAmbulanceTeam)message;

                if (!wi.getChanged().getChangedEntities().contains(mat.getAgentID()))
                    WorldUtil.reflectMessage(wi, mat);

                if (mat.getAction() == MessageAmbulanceTeam.ACTION_RESCUE)
                {
                    StandardEntity entity = wi.getEntity(mat.getTargetID());

                    if (entity != null && entity instanceof Human)
                    {
                        Human human = (Human)entity;
                        if (!human.isPositionDefined())
                            human.setPosition(mat.getPosition());
                    }
                }

                if (mat.getAction() == MessageAmbulanceTeam.ACTION_LOAD)
                {
                    StandardEntity entity = wi.getEntity(mat.getTargetID());

                    if (entity != null && entity instanceof Civilian)
                    {
                        Civilian civilian = (Civilian)entity;
                        if (!civilian.isPositionDefined())
                            civilian.setPosition(mat.getAgentID());
                    }
                }
            }

            if(message.getClass() == MessageFireBrigade.class)
            {
                MessageFireBrigade mfb = (MessageFireBrigade)message;
                if (!wi.getChanged().getChangedEntities().contains(mfb.getAgentID()))
                    WorldUtil.reflectMessage(wi, mfb);
            }

            if (message.getClass() == MessagePoliceForce.class)
            {
                MessagePoliceForce mpf = (MessagePoliceForce)message;
                if (!wi.getChanged().getChangedEntities().contains(mpf.getAgentID()))
                    WorldUtil.reflectMessage(wi, mpf);
            }
        }
    }
}
