package naito.util.passable;

import rescuecore2.misc.geometry.*;
import rescuecore2.standard.entities.Area;

public class AreaGrid
{
    private final static double AGENT_SIZE = 500;

    public static Point2D[][] create(Area area, Point2D entrancePoint, Point2D exitPoint)
    {
        java.awt.Shape shape = area.getShape();

        //java.awt.geom.AffineTransform transformer = java.awt.geom.AffineTransform.getRotateInstance(
        //    Math.toRadians(-calculateAngle(entrancePoint, exitPoint)),
        //    entrancePoint.getX(),
        //    entrancePoint.getY()
        //);
        //java.awt.Shape rotatedShape = transformer.createTransformedShape(shape);

        //java.awt.geom.Rectangle2D bound = rotatedShape.getBounds2D();
        java.awt.geom.Rectangle2D bound = shape.getBounds2D();
        double beginX = bound.getX();
        double beginY = bound.getY();
        double width  = bound.getWidth();
        double height = bound.getHeight();
        double finalX = beginX + width;
        double finalY = beginY + height;

        int wsize = ((int)(width  / AGENT_SIZE))+1;
        int hsize = ((int)(height / AGENT_SIZE))+1;

        //  init grid array
        Point2D[][] grid = new Point2D[wsize][hsize];
        for (int w=0; w<wsize; w++) for (int h=0; h<hsize; h++)
        {
            double x = (w == (wsize-1)) ? finalX : beginX+(w*AGENT_SIZE);
            double y = (h == (hsize-1)) ? finalY : beginY+(h*AGENT_SIZE);
            //if (w == 0) x+=5.0;
            //if (h == 0) y+=5.0;

            //java.awt.geom.Point2D rotatedPoint =
            //    new java.awt.geom.Point2D.Double(x, y);
            //java.awt.geom.Point2D rightPoint =
            //    new java.awt.geom.Point2D.Double();
            //try
            //{
            //    transformer.inverseTransform(rotatedPoint, rightPoint);
            //}
            //catch (java.awt.geom.NoninvertibleTransformException e)
            //{
            //    System.out.println("AreaGrid: NoninvertibleTransformException");
            //}

            //grid[w][h] = new Point2D(rightPoint.getX(), rightPoint.getY());
            grid[w][h] = new Point2D(x, y);
        }

        //  exclusion out of the area
        for (int w=0; w<wsize; w++) for (int h=0; h<hsize; h++)
        {
            Point2D point = grid[w][h];
            if (!shape.contains( point.getX(), point.getY() ))
            {
                grid[w][h] = null;
            }
        }

        return grid;
    }

    public static Point2D[][] create(Area area, Line2D entranceLine, Line2D exitLine)
    {
        Point2D entrancePoint = GeometryMath.center(entranceLine);
        Point2D exitPoint = GeometryMath.center(exitLine);

        return create(area, entrancePoint, exitPoint);
    }

    public static Point2D[][] create(Area area, Point2D entrancePoint, Line2D exitLine)
    {
        Point2D exitPoint = GeometryMath.center(exitLine);

        return create(area, entrancePoint, exitPoint);
    }

    public static Point2D[][] create(Area area, Line2D entranceLine, Point2D exitPoint)
    {
        Point2D entrancePoint = GeometryMath.center(entranceLine);

        return create(area, entrancePoint, exitPoint);
    }

    private static double calculateAngle(Point2D origin, Point2D end)
    {
        double angle = GeometryTools2D.getLeftAngleBetweenLines(
            new Line2D(origin, new Point2D(origin.getX()+100.0, origin.getY())),
            new Line2D(origin, end)
        );
        return angle;
    }
}
