package naito.util.passable;

import naito.util.passable.astar.AStar;
import rescuecore2.misc.geometry.*;
import rescuecore2.standard.entities.Edge;
import rescuecore2.standard.entities.Area;
import rescuecore2.standard.entities.Blockade;
import adf.agent.info.WorldInfo;
import naito.util.passable.astar.Index;
import naito.util.passable.astar.Node;

import java.util.*;
import java.util.stream.*;

public class SearchPoliceClearLine
{
    private static final double AGENT_SIZE = 500;

    public List<List<Line2D>> search(Area area, Line2D in, Line2D out, WorldInfo wi)
    {
        //System.out.println(this.getClass().getSimpleName() + ": target=" + area.getID());
        //  @STEP1 make grid of area
        Point2D[][] grid = AreaGrid.create(area, in, out);

        //  @STEP2 astar for grid of area
        AStar<Point2D> astar = new AStar<>(grid);
        //astar.print();

        //  @STEP3  lower node's step in area's blockades
        List<Blockade> blockades = AreaUtil.toBlockades(area, wi);
        this.lowerNodePriorityInShapes(
            astar.getNodes(),
            AreaUtil.toShapes(blockades)
        );

        //  @STEP4 astar : set begin and goal index
        List<Index> beginIndexes = this.findIndexesOfContiguous(in , grid);
        List<Index>  goalIndexes = this.findIndexesOfContiguous(out, grid);
        if (beginIndexes.size() == 0)
        {
            //System.out.println(this.getClass().getSimpleName() + ": beginIndex error");
            return null;
        }
        if ( goalIndexes.size() == 0)
        {
            //System.out.println(this.getClass().getSimpleName() + ":  goalIndex error");
            return null;
        }

        astar.setBegin(
            this.extractIndexOfHighestPriorityAndCloser(
                GeometryMath.center(in),
                beginIndexes,
                astar.getNodes()
            )
        );
        astar.setGoal(
            this.extractIndexOfHighestPriorityAndCloser(
                GeometryMath.center(out),
                goalIndexes,
                astar.getNodes()
            )
        );


        //  @STEP5 astar : calc
        List<Node<Point2D>> nodePath = astar.calc().getResultNodes();
        if (nodePath.size() == 0)
        {
            //System.out.println(this.getClass().getSimpleName() + ": can't detect path");
            return null;
        }


        //  @STEP6 extract blockade's line
        List< List<Line2D> > clearLines =
            this.extractContinualPointsInBlockade(nodePath)
                .stream()
                .map(l -> this.continualPointsToContinualLines(l))
                .collect(Collectors.toList());

        return clearLines;
    }

    public List<List<Line2D>> search(Area area, Area in, Area out, WorldInfo wi)
    {
        Edge  inEdge = this.findDuplicateEdge(area, in);
        if ( inEdge == null) return null;

        Edge outEdge = this.findDuplicateEdge(area, out);
        if (outEdge == null) return null;

        //System.out.println(this.getClass().getSimpleName() + ": exit=" + out.getID());

        return
            this.search(area, inEdge.getLine(), outEdge.getLine(), wi);
    }

    public List<List<Line2D>> search(Area area, Area in, Line2D out, WorldInfo wi)
    {
        Edge  inEdge = this.findDuplicateEdge(area, in);
        if ( inEdge == null) return null;

        return
            this.search(area, inEdge.getLine(), out, wi);
    }

    public List<List<Line2D>> search(Area area, Point2D inPoint, Line2D out, WorldInfo wi)
    {
        //  @STEP1 make grid of area
        Point2D[][] grid = AreaGrid.create(area, inPoint, out);

        //  @STEP2 astar for grid of area
        AStar<Point2D> astar = new AStar<>(grid);

        //  @STEP3  lower node's step in area's blockades
        List<Blockade> blockades = AreaUtil.toBlockades(area, wi);
        this.lowerNodePriorityInShapes(
            astar.getNodes(),
            AreaUtil.toShapes(blockades)
        );

        //  @STEP4 astar : set begin and goal index
        List<Index>  goalIndexes = this.findIndexesOfContiguous(out, grid);
        if ( goalIndexes.size() == 0)
        {
            //System.out.println(this.getClass().getSimpleName() + ":  goalIndex error");
            return null;
        }

        astar.setBegin(
            this.extractIndexOfClosest(
                inPoint,
                grid
            )
        );
        astar.setGoal(
            this.extractIndexOfHighestPriorityAndCloser(
                GeometryMath.center(out),
                goalIndexes,
                astar.getNodes()
            )
        );

        //  @STEP5 astar : calc
        List<Node<Point2D>> nodePath = astar.calc().getResultNodes();
        if (nodePath.size() == 0)
        {
            //System.out.println(this.getClass().getSimpleName() + ": can't detect path");
            return null;
        }


        //  @STEP6 extract blockade's line
        List< List<Line2D> > clearLines =
            this.extractContinualPointsInBlockade(nodePath)
                .stream()
                .map(l -> this.continualPointsToContinualLines(l))
                .collect(Collectors.toList());

        return clearLines;
    }

    public List<List<Line2D>> search(Area area, Point2D inPoint, Area out, WorldInfo wi)
    {
        Edge outEdge = this.findDuplicateEdge(area, out);
        if (outEdge == null)
        {
            //System.out.println("outEdge is null");
            return null;
        }

        return
            this.search(area, inPoint, outEdge.getLine(), wi);
    }

    public List<List<Line2D>> search(Area area, Line2D in, Point2D outPoint, WorldInfo wi)
    {
        //  @STEP1 make grid of area
        Point2D[][] grid = AreaGrid.create(area, in, outPoint);

        //  @STEP2 astar for grid of area
        AStar<Point2D> astar = new AStar<>(grid);

        //  @STEP3  lower node's step in area's blockades
        List<Blockade> blockades = AreaUtil.toBlockades(area, wi);
        this.lowerNodePriorityInShapes(
            astar.getNodes(),
            AreaUtil.toShapes(blockades)
        );

        //  @STEP4 astar : set begin and goal index
        List<Index> beginIndexes = this.findIndexesOfContiguous(in , grid);
        if (beginIndexes.size() == 0)
        {
            //System.out.println(this.getClass().getSimpleName() + ": beginIndex error");
            return null;
        }

        astar.setBegin(
            this.extractIndexOfHighestPriorityAndCloser(
                GeometryMath.center(in),
                beginIndexes,
                astar.getNodes()
            )
        );
        astar.setGoal(
            this.extractIndexOfClosest(
                outPoint,
                grid
            )
        );


        //  @STEP5 astar : calc
        List<Node<Point2D>> nodePath = astar.calc().getResultNodes();
        if (nodePath.size() == 0)
        {
            //System.out.println(this.getClass().getSimpleName() + ": can't detect path");
            return null;
        }


        //  @STEP6 extract blockade's line
        List< List<Line2D> > clearLines =
            this.extractContinualPointsInBlockade(nodePath)
                .stream()
                .map(l -> this.continualPointsToContinualLines(l))
                .collect(Collectors.toList());

        return clearLines;
    }

    public List<List<Line2D>> search(Area area, Area in, Point2D outPoint, WorldInfo wi)
    {
        Edge  inEdge = this.findDuplicateEdge(area, in);
        if ( inEdge == null) return null;

        return
            this.search(area, inEdge.getLine(), outPoint, wi);
    }

    private Edge findDuplicateEdge(Area area1, Area area2)
    {
        List<Edge> edges1 = area1.getEdges().stream()
            .filter(Edge::isPassable)
            .collect(Collectors.toList());

        List<Edge> edges2 = area2.getEdges().stream()
            .filter(Edge::isPassable)
            .collect(Collectors.toList());

        for (Edge edge1 : edges1) for (Edge edge2 : edges2)
        {
            Line2D line1 = edge1.getLine();
            Line2D line2 = edge2.getLine();
            if (this.isEqual(line1, line2))
                return edge1;
        }

        return null;
    }

    private boolean isEqual(Line2D line1, Line2D line2)
    {
        Point2D origin1 = line1.getOrigin();
        Point2D origin2 = line2.getOrigin();
        Point2D end1 = line1.getEndPoint();
        Point2D end2 = line2.getEndPoint();
        if (origin1.equals(origin2) && end1.equals(end2))
            return true;

        if (origin1.equals(end2) && origin2.equals(end1))
            return true;

        return false;
    }


    private void lowerNodePriorityInShapes(
        Node<Point2D>[][]    nodes,
        List<java.awt.Shape> shapes
    ){
        int maxX = nodes.length;
        int maxY = nodes[0].length;

        for (int x=0; x<maxX; x++) for (int y=0; y<maxY; y++)
        {
            Node<Point2D> node = nodes[x][y];
            if (node == null) continue;
            Point2D point = node.getContent();

            for (java.awt.Shape shape : shapes)
                if ( shape.contains(point.getX(), point.getY()) )
                {
                    node.setStep(node.getStep()+1);
                    break;
                }
        }

        return;
    }

    private List<Index> findIndexesOfContiguous(
        Line2D      line,
        Point2D[][] points
    ){
        int maxX = points.length;
        int maxY = points[0].length;

        List<Index> indexes = new ArrayList<>();

        for (int x=0; x<maxX; x++) for (int y=0; y<maxY; y++)
        {
            Point2D point = points[x][y];
            if (point == null) continue;

            //if (GeometryMath.distance(line, point) < AGENT_SIZE*1.5)
            if (GeometryMath.distance(line, point) < AGENT_SIZE*1.5)
                indexes.add( new Index(x, y) );
        }

        return indexes;
    }

    private Index extractIndexOfHighestPriorityAndCloser(
        Point2D           point,
        List<Index>       candidacies,
        Node<Point2D>[][] nodes
    ){
        Index  bestIndex   = null;
        double minDistance =  Double.MAX_VALUE;
        double minStep     = Integer.MAX_VALUE; // priority

        for (Index index: candidacies)
        {
            Node<Point2D> node = index.toElement(nodes);
            int    step     = node.getStep();
            double distance = GeometryTools2D.getDistance(
                point,
                node.getContent()
            );

            if (step <  minStep ||
               (step == minStep && distance < minDistance)
            ){
                bestIndex   = index;
                minStep     = step;
                minDistance = distance;
            }
        }

        return bestIndex;
    }

    private Index extractIndexOfClosest(
        Point2D     point,
        Point2D[][] points
    ){
        int maxX = points.length;
        int maxY = points[0].length;

        Index  bestIndex   = null;
        double minDistance = Double.MAX_VALUE;

        for (int x=0; x<maxX; x++) for (int y=0; y<maxY; y++)
        {
            if (points[x][y] == null)
                continue;

            double distance = GeometryTools2D.getDistance(
                point,
                points[x][y]
            );

            if (distance < minDistance)
            {
                bestIndex   = new Index(x, y);
                minDistance = distance;
            }
        }

        return bestIndex;
    }

    private List<List<Point2D>> extractContinualPointsInBlockade(
        List<Node<Point2D>> nodes
    ){
        List<List<Point2D>> continualPointsList = new ArrayList<>();
        List<Point2D> presentPoints = new ArrayList<>();

        for (Node<Point2D> node: nodes)
        {
            if (node.getStep() != 2)
            {
                continualPointsList.add(presentPoints);
                presentPoints = new ArrayList<>();
                continue;
            }

            presentPoints.add(node.getContent());
        }

        continualPointsList.add(presentPoints);

        return continualPointsList.stream()
            .filter(l -> l.size() > 0)
            .collect(Collectors.toList());
    }

    private List<Line2D> continualPointsToContinualLines(
        List<Point2D> points
    ){
        List<Line2D> continualLines = new ArrayList<>();
        Line2D presentLine = null;

        if (points.size() == 1)
        {
            Line2D line = new Line2D(points.get(0), points.get(0));
            continualLines.add(line);
            return continualLines;
        }

        for (Line2D line : GeometryTools2D.pointsToLines(points))
        {
            if (presentLine == null)
            {
                presentLine = line;
                continue;
            }

            if (!GeometryTools2D.parallel(line, presentLine))
            {
                continualLines.add(presentLine);
                presentLine = line;
                continue;
            }

            presentLine = new Line2D(
                presentLine.getOrigin(),
                line.getEndPoint()
            );
        }

        if (presentLine != null)
            continualLines.add(presentLine);

        return continualLines;
    }
}
