package naito.util.passable.astar;

public class Node<T>
{
    private T       content;
    private State   state;
    private int     step;
    private int     costToSelf;
    private int     costToGoalX;
    private int     costToGoalY;
    private int     score;
    private Index   footpoint;

    public Node(T content)
    {
        this.content = content;
        this.state = State.None;
        this.step        = 1;
        this.costToSelf  = 0;
        this.costToGoalX = 0;
        this.costToGoalY = 0;
        this.score       = 0;
    }

    public boolean isCloser(Node<T> node)
    {
        if (node.getScore() < this.getScore())
            return true;

        if (node.getScore() != this.getScore())
            return false;

        if (node.getCostToGoalX() < this.getCostToGoalX() ||
            node.getCostToGoalY() < this.getCostToGoalY()
        )   return true;

        return false;
    }

    public T getContent()
    {
        return this.content;
    }

    public State getState()
    {
        return this.state;
    }

    public int getCostToSelf()
    {
        return this.costToSelf;
    }

    public int getCostToGoal()
    {
        return (this.costToGoalX >= this.costToGoalY)
            ?   this.costToGoalX
            :   this.costToGoalY
            ;
    }

    public int getCostToGoalX()
    {
        return this.costToGoalX;
    }

    public int getCostToGoalY()
    {
        return this.costToGoalY;
    }

    public int getScore()
    {
        return this.getCostToSelf() + this.getCostToGoal();
    }

    public void setState(State state)
    {
        this.state = state;
    }

    public void setCostToSelf(int cost)
    {
        this.costToSelf = cost;
    }

    public void setCostToGoal(int costX, int costY)
    {
        this.costToGoalX = costX;
        this.costToGoalY = costY;
    }

    public Index getFootpoint()
    {
        return this.footpoint;
    }

    public void setFootpoint(Index footpoint)
    {
        this.footpoint = footpoint;
    }

    public int getStep()
    {
        return this.step;
    }

    public void setStep(int step)
    {
        this.step = step;
    }
}
